"""This file defines a representation of the Kepler Field of View, `KeplerFov`.
"""
import numpy as np

try:
    import matplotlib.pyplot as mp
    import matplotlib
except ImportError:
    pass

from . import logging
from . import projection as proj
from . import rotate2 as r
from . import greatcircle as gcircle
from . import definefov

from . import DEFAULT_PADDING


"""
According to Instrument Handbook page 47, mod3 is in the +y direction
coordinates (+x is in the direction of the telescope pointing).

According to Flight Segment Users' Manual, p92 the sun shield is in
the +y  directions.

So in the model used in KeplerFov, at ra,dec=0, a roll angle of 0
means pointing the solar array due North.

In two-wheel mode, we will want the roll angle equal to the solar
angle.
"""


def getFovAngleFromSpacecraftRoll(yaxis_deg):
    """The y-axis vector (perpendicular to the solar arrays)
    lies 193 degrees clockwise of the angle
    from the centre of the FOV to the centre of mod 3

    As a diagram:

        4  3  2
     9  8  7  6  5
    14 13 12 11 10
          \
           \
            \
             \ y-axis of S/C in this direction

   This function converts from angles relative to spacecraft y-axis
   to angles relative to the FOV
   """
    return yaxis_deg + 13.0 + 180 - 90


def getSpacecraftRollAngleFromFovAngle(fovAngle_deg):
    """See notes on getFovAngleFromSpacecraftYAxisAngle()"""

    return fovAngle_deg - 13.0 - 180 + 90


class KeplerFov():
    def __init__(self, ra_deg, dec_deg, roll_deg):
        """
        A representation of the Kepler Field of View designed
        for planning target observations.

        Inputs:
        ra_deg, dec_deg (floats) The direction to point the boresight
                        of the telescope. Note that while any
                        legal ra/dec is accepted by this class,
                        allowed values in the two wheel
                        mission are tightly constrained. Values
                        are in degrees

        roll_deg        Roll of the boresight. a roll of zero
                        orients the FOV so that mod 3 is due North.
                        Use getFovAngleFromSpacecraftRoll()
                        to convert from spacecraft roll values


        Values for the prime mission are:
        ra_deg = 290.66666667
        dec_deg = +44.5
        rollAngle_deg = 33.0 + n*90, where n is the season number
        """
        # default map is set by setPointing()
        # This is used for calculations of where objects lie within
        # a channel, and is always a Gnomic projection centred on
        # the boresight
        self.defaultMap = None

        self.brokenChannels = [5,6,7,8, 17,18,19,20]
        self.plateScale_arcsecPerPix = 3.98

        self.mods = list(range(1, 25))
        # Remove the 4 FGS mods
        self.mods.pop(0)
        self.mods.pop(3)
        self.mods.pop(-4)

        # Relative vectors to the module corners.
        # If the spacecraft was pointed at (ra, dec) = (0,0) with mod3
        # pointed north, r.raDecFromVec() of these values would give
        # The ra and decs of the corners of the modules.
        self.origin = definefov.loadOriginVectors()

        self.ra0_deg = ra_deg
        self.dec0_deg = dec_deg
        self.roll0_deg = roll_deg

        self.currentRaDec = None
        self.setPointing(ra_deg, dec_deg, roll_deg)

    ###
    # Code related to pointing the spacecraft
    ###

    def getOrigin(self, cartesian=False):
        """Return the ra/decs of the channel corners if the S/C
        is pointed at the origin (ra,dec = 0,0)

        Inputs:
        cartesian   (bool) If True, return each channel corner
                    as a unit vector

        Returns:
        A 2d numpy array. Each row represents a channel corner
        The columns are module, output, channel, ra, dec

        If cartestian is True, ra, and dec are replaced by the
        coordinates of a 3 vector
        """
        out = self.origin.copy()

        if cartesian is False:
            out = self.getRaDecs(out)
        return out

    def setPointing(self, ra_deg, dec_deg, roll_deg):
        t = self.computePointing(ra_deg, dec_deg, roll_deg)
        self.currentRaDec = t
        self.defaultMap = proj.Gnomic(ra_deg, dec_deg)

        self.ra0_deg = ra_deg
        self.dec0_deg = dec_deg
        self.roll0_deg = roll_deg

    def computePointing(self, ra_deg, dec_deg, roll_deg, cartesian=False):
        """Compute a pointing model without changing the internal object pointing"""
        # Roll FOV
        Rrotate = r.rotateInXMat(roll_deg)  # Roll

        # Slew from ra/dec of zero
        Ra = r.rightAscensionRotationMatrix(ra_deg)
        Rd = r.declinationRotationMatrix(dec_deg)
        Rslew = np.dot(Ra, Rd)

        R = np.dot(Rslew, Rrotate)

        slew = self.origin*1
        for i, row in enumerate(self.origin):
            slew[i, 3:6] = np.dot(R, row[3:6])

        if cartesian is False:
            slew = self.getRaDecs(slew)
        return slew

    def getRaDecs(self, mods):
        """Internal function converting cartesian coords to
        ra dec"""
        raDecOut = np.empty( (len(mods), 5))
        raDecOut[:,0:3] = mods[:,0:3]

        for i, row in enumerate(mods):
            raDecOut[i, 3:5] = r.raDecFromVec(row[3:6])
        return raDecOut

    def getCoordsOfChannelCorners(self):
        """Get ra/decs of corners of channels.

        Input:
        (none)

        Returns:
        A 2d numpy array.

        Each row represents a single corner of a channel.
        The columns are:
        module, output, channel, ra (degrees), dec (degrees)

        Note that the locations of the FGS channels are
        included in this output. FGS channels are 85-88
        inclusive
        """
        return self.currentRaDec

    def getBoresight(self):
        # Trap potential negative values for ra
        alpha = self.ra0_deg
        if alpha < 0:
            alpha += 360.

        return alpha, self.dec0_deg, self.roll0_deg


    ###
    # Sky -> pixel code
    ###

    def isOnSilicon(self, ra_deg, dec_deg, padding_pix=DEFAULT_PADDING):
        """Returns True if the given location is observable with a science CCD.

        Parameters
        ----------
        ra_deg : float
            Right Ascension (J2000) in decimal degrees.

        dec_deg : float
            Declination (J2000) in decimal degrees.

        padding : float
            Objects <=  this many pixels off the edge of a channel are counted
            as inside.  This allows one to compensate for the slight
            inaccuracy in `K2fov` that results from e.g. the lack of optical
            distortion modeling.
        """
        ch, col, row = self.getChannelColRow(ra_deg, dec_deg)
        # Modules 3 and 7 are no longer operational
        if ch in self.brokenChannels:
            return False
        # K2fov encodes the Fine Guidance Sensors (FGS) as
        # "channel" numbers 85-88; they are not science CCDs.
        if ch > 84:
            return False
        return self.colRowIsOnSciencePixel(col, row, padding_pix)

    def getChannelColRow(self, ra, dec, wantZeroOffset=False,
                         allowIllegalReturnValues=True):
        """Returns (channel, column, row) given an (ra, dec) coordinate.

        Returns (0, 0, 0) or a ValueError if the coordinate is not on silicon.
        """
        try:
            ch = self.pickAChannel(ra, dec)
        except ValueError:
            logger.warning("WARN: %.7f %.7f not on any channel" % (ra, dec))
            return (0, 0, 0)

        col, row = self.getColRowWithinChannel(ra, dec, ch, wantZeroOffset,
                                               allowIllegalReturnValues)
        return (ch, col, row)

    def pickAChannel(self, ra_deg, dec_deg):
        """Returns the channel number closest to a given (ra, dec) coordinate.
        """
        # Could improve speed by doing this in the projection plane
        # instead of sky coords
        cRa = self.currentRaDec[:, 3]  # Ra of each channel corner
        cDec = self.currentRaDec[:, 4]  # dec of each channel corner

        dist = cRa * 0
        for i in range(len(dist)):
            dist[i] = gcircle.sphericalAngSep(cRa[i], cDec[i], ra_deg, dec_deg)

        i = np.argmin(dist)
        return self.currentRaDec[i, 2]

    def getColRowWithinChannel(self, ra, dec, ch, wantZeroOffset=False,
                               allowIllegalReturnValues=True):
        """Returns (col, row) given a (ra, dec) coordinate and channel number.
        """
        # How close is a given ra/dec to the origin of a KeplerModule?
        x, y = self.defaultMap.skyToPix(ra, dec)
        kepModule = self.getChannelAsPolygon(ch)
        r = np.array([x[0],y[0]]) - kepModule.polygon[0, :]

        v1 = kepModule.polygon[1, :] - kepModule.polygon[0, :]
        v3 = kepModule.polygon[3, :] - kepModule.polygon[0, :]

        # Divide by |v|^2 because you're normalising v and r
        colFrac = np.dot(r, v1) / np.linalg.norm(v1)**2
        rowFrac = np.dot(r, v3) / np.linalg.norm(v3)**2

        # This is where it gets a little hairy. The channel "corners"
        # supplied to me actually represent points 5x5 pixels inside
        # the science array. Which isn't what you'd expect.
        # These magic numbers are the pixel numbers of the corner
        # edges given in fov.txt
        col = colFrac*(1106-17) + 17
        row = rowFrac*(1038-25) + 25

        if not allowIllegalReturnValues:
            if not self.colRowIsOnSciencePixel(col, row):
                msg = "Request position %7f %.7f " % (ra, dec)
                msg += "does not lie on science pixels for channel %i " % (ch)
                msg += "[ %.1f %.1f]" % (col, row)
                raise ValueError(msg)

        # Convert from zero-offset to one-offset coords
        if not wantZeroOffset:
            col += 1
            row += 1

        return (col, row)

    def colRowIsOnSciencePixel(self, col, row, padding=DEFAULT_PADDING):
        """Is col row on a science pixel?

        Ranges taken from Fig 25 or Instrument Handbook (p50)

        Padding allows for the fact that distortion means the
        results from getColRowWithinChannel can be off by a bit.
        Setting padding > 0 means that objects that are computed
        to lie a small amount off silicon will return True.

        To be conservative, set padding to negative
        """
        if col < 12. - padding or col > 1111 + padding:
            return False

        if row < 20 - padding or row > 1043 + padding:
            return False
        return True

    def getColRowWithinFgsCh(self, ra, dec, ch, wantZeroOffset=False,
                             allowIllegalReturnValues=True):
        """How close is a given ra/dec to the origin of an FGS mod

        Returns col and row of the position.
        """
        x, y = self.defaultMap.skyToPix(ra, dec)
        kepModule = self.getChannelAsPolygon(ch)
        r = np.array([x[0],y[0]]) - kepModule.polygon[0,:]

        v1 = kepModule.polygon[1,:] - kepModule.polygon[0,:]
        v3 = kepModule.polygon[3,:] - kepModule.polygon[0,:]

        colFrac = np.dot(r, v1) / np.linalg.norm(v1)**2
        rowFrac = np.dot(r, v3) / np.linalg.norm(v3)**2

        col = colFrac*(547)
        row = rowFrac*(527)

        if not allowIllegalReturnValues:
            if not self.colRowIsOnFgsPixel(col, row):
                msg = "Request position %7f %.7f " %(ra, dec)
                msg += "does not lie on FGS pixels for channel %i " %(ch)
                msg += "[ %.1f %.1f]" %(col, row)
                raise ValueError(msg)

        # Convert from zero-offset to one-offset coords
        if not wantZeroOffset:
            col += 1
            row += 1

        return (col, row)

    def colRowIsOnFgsPixel(self, col, row, padding=-50):
        """Is col row on a science pixel?

        #See Kepler Flight Segment User's Manual (SP0039-702) \S 5.4 (p88)

        Inputs:
        col, row (floats or ints)
        padding    If padding <0, pixel must be on silicon and this many
                   pixels away from the edge of the CCD to return True

        Returns:
        boolean
        """
        if col < 12. - padding  or col > 547 + padding:
            return False

        if row < 0 - padding or row > 527 + padding :
            return False
        return True

    ###
    # Pixel --> sky
    ###
    def getRaDecForChannelColRow(self, ch, col, row, oneOffsetPixels=True):

        # To FGS channels correctly
        if ch > 84:
            return self.getRaDecForFgsChannelColRow(ch, col, row,\
                oneOffsetPixels)

        if oneOffsetPixels:
            col -= 1
            row -= 1

        # Convert col row to colFrac, rowFrac
        # See notes in getColRowWithinChannel
        colFrac = (col-17.) / (1106.-17.)
        rowFrac = (row-25.) / (1038.-25.)

        # Get basis vectors for channel. vZero is vector close
        # to readout of chip (c,r) = (0,0)
        # vCol is a vector in increasing column direction
        kepModule = self.getChannelAsPolygon(ch)
        vZero = kepModule.polygon[0,:]
        vCol = kepModule.polygon[1,:] - vZero
        vRow = kepModule.polygon[3,:] - vZero

        # Where on the projected plane does col,row lie?
        projectionXy = vZero + (colFrac*vCol) + (rowFrac*vRow)

        # Call pixToSky
        x, y = projectionXy
        a, d = self.defaultMap.pixToSky(x, y)
        return [a[0], d[0]]

    def getRaDecForFgsChannelColRow(self, ch, col, row, oneOffsetPixels=True):
        if oneOffsetPixels:
            col -= 1
            row -= 1

        # Convert col row to colFrac, rowFrac
        # Offsets here haven't been rigourously checked.
        colFrac = (col-0.) / (547.)
        rowFrac = (row-0.) / (527.)

        # Get basis vectors for channel. vZero is vector close
        # to readout of chip (c,r) = (0,0)
        # vCol is a vector in increasing column direction
        kepModule = self.getChannelAsPolygon(ch)
        vZero = kepModule.polygon[0,:]
        vCol = kepModule.polygon[1,:] - vZero
        vRow = kepModule.polygon[3,:] - vZero

        # Where on the projected plane does col,row lie?
        projectionXy = vZero + (colFrac*vCol) + (rowFrac*vRow)

        # Call pixToSky
        x, y = projectionXy
        a, d = self.defaultMap.pixToSky(x, y)
        return [a[0], d[0]]

    ###
    #  Polygon code: sky <--> pix and other functions use
    #  these polygons to represent a single channel on the FOV
    ###
    def getAllChannelsAsPolygons(self, maptype=None):
        """Return slew the telescope and return the corners of the modules
        as Polygon objects.

        If a projection is supplied, the ras and
        decs are mapped onto x, y using that projection
        """
        polyList = []
        for ch in self.origin[:, 2]:
            poly = self.getChannelAsPolygon(ch, maptype)
            polyList.append(poly)
        return polyList

    def getChannelAsPolygon(self, chNumber, maptype=None):
        if maptype is None:
            maptype = self.defaultMap

        radec = self.currentRaDec
        idx = np.where(radec[:, 2].astype(np.int) == chNumber)[0]

        if not np.any(idx):
            raise ValueError("%i is not a valid channel number" % (chNumber))

        x, y = maptype.skyToPix(radec[idx, 3], radec[idx, 4])
        return KeplerModOut(chNumber, x=x, y=y)


    ###
    # Plotting code
    ###

    def plotPointing(self, maptype=None, colour='b', mod3='r', showOuts=True, **kwargs):
        """Plot the FOV
        """

        if maptype is None:
            maptype=self.defaultMap

        radec = self.currentRaDec
        for ch in radec[:,2][::4]:

            idx = np.where(radec[:,2].astype(np.int) == ch)[0]
            idx = np.append(idx, idx[0])  #% points to draw a box

            c = colour
            if ch in self.brokenChannels:
                c = mod3
            maptype.plot(radec[idx, 3], radec[idx, 4], '-', color=c, **kwargs)
            #Show the origin of the col and row coords for this ch
            if showOuts:
                maptype.plot(radec[idx[0], 3], radec[idx[0],4], 'o', color=c)


    def plotOutline(self, maptype=None, colour='#AAAAAA', **kwargs):
        """Plot an outline of the FOV.
        """

        if maptype is None:
            maptype=self.defaultMap

        xarr = []
        yarr = []
        radec = self.currentRaDec
        for ch in [20,4,11,28,32, 71,68, 84, 75, 60, 56, 15 ]:
            idx = np.where(radec[:,2].astype(np.int) == ch)[0]
            idx = idx[0]    #Take on the first one
            x, y = maptype.skyToPix(radec[idx][3], radec[idx][4])
            xarr.append(x)
            yarr.append(y)

        verts = np.empty( (len(xarr), 2))
        verts[:,0] = xarr
        verts[:,1] = yarr

        #There are two ways to specify line colour
        ec = kwargs.pop('ec', "none")
        ec = kwargs.pop('edgecolor', ec)
        p = matplotlib.patches.Polygon(verts, fill=True, ec=ec, fc=colour, **kwargs)
        mp.gca().add_patch(p)


    def plotSpacecraftYAxis(self, maptype=None):
        """Plot a line pointing in the direction of the spacecraft
        y-axis (i.e normal to the solar panel
        """

        if maptype is None:
            maptype=self.defaultMap
        #Plot direction of spacecraft +y axis. The subtraction of
        #90 degrees accounts for the different defintions of where
        #zero roll is.
        yAngle_deg = getSpacecraftRollAngleFromFovAngle(self.roll0_deg)
        yAngle_deg -=90

        a,d = gcircle.sphericalAngDestination(self.ra0_deg, self.dec0_deg, -yAngle_deg, 12.0)
        x0, y0 = maptype.skyToPix(self.ra0_deg, self.dec0_deg)
        x1, y1 = maptype.skyToPix(a, d)
        mp.plot([x0, x1], [y0, y1], 'k-')


    def plotChIds(self, maptype=None, modout=False):
        """Print the channel numbers on the plotting display"""
        if maptype is None:
            maptype = self.defaultMap

        polyList = self.getAllChannelsAsPolygons(maptype)
        for p in polyList:
            p.identifyModule(modout=modout, maptype=maptype)


    def getWcsForChannel1(self, ch):
        crpix =np.array( [500, 500])    #Rough guess at centre
        a,d = self.getRaDecForChannelColRow(ch, crpix[0], crpix[1])
        crval = np.array([a,d])

        #Get rotation of channel relative to FOV
        kepModule = self.getChannelAsPolygon(ch)
        vZero =  kepModule.polygon[0,:]
        vCol = kepModule.polygon[1,:] - vZero
        vRow = kepModule.polygon[3,:] - vZero
        ang_rad = np.arctan2(vCol[1], vCol[0])
        #ang_rad -= np.radians(.2308)   #Debugging code

        if np.cross(vCol, vRow) >= 0:
            sign = +1
        else:
            sign = -1

        CD = np.empty( (2,2))
        CD[0,0] = np.cos(ang_rad)
        CD[0,1] = np.sin(ang_rad)
        CD[1,0] = -np.sin(ang_rad)
        CD[1,1] = np.cos(ang_rad)

        if sign < 0:
            CD[1,:] *= -1

        CD *= self.plateScale_arcsecPerPix/3600.

        return crval, crpix, CD




###############################################
# Polygon and KepModule code
################################################

class Polygon():
    def __init__(self, x=None, y=None, pointList=None):
        """

        Input
        pointList   A list of (x,y) pairs. Eg
                    [ (0,0), (1,0), (0,1), (1.1)]

                    The edges of the polygon join adjacent elements
                    of this list, so the order matters. The last
                    point is assumed to connect to the first point.
        """

        if x is not None and y is not None:
            pointList = []
            for xi, yi in zip(x, y):
                pointList.append( (xi, yi))

        if pointList is None:
            raise ValueError("Must supply x,y or pointList")

        self.polygon = np.array(pointList)


    def __str__(self):
        return self.polygon.__str__()

    def __repr__(self):
        return self.polygon.__repr__()

    def isPointInside(self, xp, yp):
        """Is the given point inside the polygon?

        Input:
        polygon (nx2 numpy array). polygon[i] = [x, y] coords of
                a vertex of a polygon
        point   (1x2) numpy array) x,y coords of the point we wish
                to determine if it's in the polygon or not.

        Returns true/ false
        Does this work in >2 dimensions? Probably, with a little bit
        of work
        """

        point = np.array([xp, yp]).transpose()
        polygon = self.polygon
        numVert, numDim = polygon.shape

        #Subtract each point from the previous one.
        polyVec = np.roll(polygon, -1, 0) - polygon
        #Get the vector from each vertex to the given point
        pointVec = point - polygon

        crossProduct = np.cross(polyVec, pointVec)

        if np.all(crossProduct < 0) or np.all(crossProduct > 0):
            return True
        return False

    def draw(self, **kwargs):
        ax = mp.gca()
        shape = matplotlib.patches.Polygon(self.polygon, **kwargs)
        ax.add_artist(shape)


class KeplerModOut(Polygon):
    def __init__(self, channel, x=None, y=None, pointList=None):
        Polygon.__init__(self, x,y,pointList)
        self.channel = channel

    def getChannel(self):
        return self.channel


    def identifyModule(self, modout=False):
        x,y = np.mean(self.polygon, 0)

        if modout:
            modout = modOutFromChannel(self.channel)
            mp.text(x, y, "%i-%i" %(modout[0], modout[1]), fontsize=8, ha="center", clip_on=True)
        else:
            mp.text(x,y, "%i" %(self.channel), fontsize=8, ha="center", clip_on=True)




#########################################################
#  channel <--> mod out
#########################################################

def channelFromModOut(mod, out):
    lookup = loadChannelModOutLookup()
    return lookup[mod, out]


def modOutFromChannel(ch):
    lookup = loadChannelModOutLookup()
    idx = lookup == ch
    idx[:,0] = False

    if not np.any(idx):
        raise ValueError("Illegal channel request")

    if np.sum(idx) > 1:
        raise ValueError("Channel number begins at 1, not zero")

    modout = np.where(idx)
    mod = modout[0][0]
    out = modout[1][0]
    return (mod, out)


def loadChannelModOutLookup():
    lookup = np.array( [ \
       [ 0,    0,    0,    0,    0], \
       [ 1,   85,    0,    0,    0], \
       [ 2,    1,    2,    3,    4], \
       [ 3,    5,    6,    7,    8], \
       [ 4,    9,   10,   11,   12], \
       [ 5,   86,    0,    0,    0], \
       [ 6,   13,   14,   15,   16], \
       [ 7,   17,   18,   19,   20], \
       [ 8,   21,   22,   23,   24], \
       [ 9,   25,   26,   27,   28], \
       [10,   29,   30,   31,   32], \
       [11,   33,   34,   35,   36], \
       [12,   37,   38,   39,   40], \
       [13,   41,   42,   43,   44], \
       [14,   45,   46,   47,   48], \
       [15,   49,   50,   51,   52], \
       [16,   53,   54,   55,   56], \
       [17,   57,   58,   59,   60], \
       [18,   61,   62,   63,   64], \
       [19,   65,   66,   67,   68], \
       [20,   69,   70,   71,   72], \
       [21,   87,    0,    0,    0], \
       [22,   73,   74,   75,   76], \
       [23,   77,   78,   79,   80], \
       [24,   81,   82,   83,   84], \
       [25,   88,    0,    0,    0], \
       ])

    return lookup


#####################################################################
#####################################################################
#####################################################################

#def getRaDecOut(vectors):
    #raDecOut = np.empty( (len(vectors), 2))
    #for i, row in enumerate(vectors):
        #raDecOut[i] = r.raDecFromVec(row)
    #return raDecOut



"""Implements the `K2findCampaigns` tools.

These tools allow to check whether a position or object is visible in any
campaign.  This complements K2onSilicon, which only determines whether an
object is on silicon for a single campaign.
"""
import argparse
import numpy as np

from . import fields
from . import logger
from . import Highlight
from .K2onSilicon import (parse_file, onSiliconCheck)


def findCampaigns(ra, dec):
    """Returns a list of the campaigns that cover a given position.

    Parameters
    ----------
    ra, dec : float, float
        Position in decimal degrees (J2000).

    Returns
    -------
    campaigns : list of int
        A list of the campaigns that cover the given position.
    """
    # Temporary disable the logger to avoid the preliminary field warnings
    logger.disabled = True
    campaigns_visible = []
    for c in fields.getFieldNumbers():
        fovobj = fields.getKeplerFov(c)
        if onSiliconCheck(ra, dec, fovobj):
            campaigns_visible.append(c)
    # Re-enable the logger
    logger.disabled = True
    return campaigns_visible


def findCampaignsByName(target):
    """Returns a list of the campaigns that cover a given target.

    Parameters
    ----------
    target : str
        Name of the celestial object.

    Returns
    -------
    campaigns : list of int
        A list of the campaigns that cover the given target name.

    ra, dec : float, float
        Resolved coordinates in decimal degrees (J2000).

    Exceptions
    ----------
    Raises an ImportError if AstroPy is not installed.
    Raises a ValueError if `name` cannot be resolved to coordinates.
    """
    # Is AstroPy (optional dependency) installed?
    try:
        from astropy.coordinates import SkyCoord
        from astropy.coordinates.name_resolve import NameResolveError
    except ImportError:
        logger.error('AstroPy needs to be installed for this feature.')
        sys.exit(1)
    # Translate the target name into celestial coordinates
    try:
        crd = SkyCoord.from_name(target)
    except NameResolveError:
        raise ValueError('Could not find coordinates '
                         'for target "{0}".'.format(target))
    # Find the campaigns with visibility
    return findCampaigns(crd.ra.deg, crd.dec.deg), crd.ra.deg, crd.dec.deg


def save_context_plots(ra, dec, targetname=""):
    from . import plot
    output_fn = "K2findCampaigns.png"
    print("Writing {0}".format(output_fn))
    myplot = plot.create_context_plot(ra, dec, name=targetname)
    myplot.fig.savefig(output_fn, dpi=300)

    output_fn = "K2findCampaigns-zoom.png"
    print("Writing {0}".format(output_fn))
    myplot = plot.create_context_plot_zoomed(ra, dec, name=targetname)
    myplot.fig.savefig(output_fn, dpi=300)


def K2findCampaigns_main(args=None):
    """Exposes K2findCampaigns to the command line."""
    parser = argparse.ArgumentParser(
                    description="Check if a celestial coordinate is "
                                "(or was) observable by any past or future "
                                "observing campaign of NASA's K2 mission.")
    parser.add_argument('ra', nargs=1, type=float,
                        help="Right Ascension in decimal degrees (J2000).")
    parser.add_argument('dec', nargs=1, type=float,
                        help="Declination in decimal degrees (J2000).")
    parser.add_argument('-p', '--plot', action='store_true',
                        help="Produce a plot showing the target position "
                             "with respect to all K2 campaigns.")
    args = parser.parse_args(args)
    ra, dec = args.ra[0], args.dec[0]
    campaigns = findCampaigns(ra, dec)
    # Print the result
    if len(campaigns):
        print(Highlight.GREEN + "Success! The target is on silicon "
              "during K2 campaigns {0}.".format(campaigns) + Highlight.END)
    else:
        print(Highlight.RED + "Sorry, the target is not on silicon "
              "during any K2 campaign." + Highlight.END)
    # Make a context plot if the user requested so
    if args.plot:
        save_context_plots(ra, dec, "Your object")


def K2findCampaigns_byname_main(args=None):
    """Exposes K2findCampaigns to the command line."""
    parser = argparse.ArgumentParser(
                    description="Check if a target is "
                                "(or was) observable by any past or future "
                                "observing campaign of NASA's K2 mission.")
    parser.add_argument('name', nargs=1, type=str,
                        help="Name of the object.  This will be passed on "
                             "to the CDS name resolver "
                             "to retrieve coordinate information.")
    parser.add_argument('-p', '--plot', action='store_true',
                        help="Produce a plot showing the target position "
                             "with respect to all K2 campaigns.")
    args = parser.parse_args(args)
    targetname = args.name[0]
    campaigns, ra, dec = findCampaignsByName(targetname)
    # Print the result
    if len(campaigns):
        print(Highlight.GREEN +
              "Success! {0} is on silicon ".format(targetname) +
              "during K2 campaigns {0}.".format(campaigns) +
              Highlight.END)
    else:
        print(Highlight.RED + "Sorry, {} is not on silicon "
              "during any K2 campaign.".format(targetname) + Highlight.END)
    # Make a context plot if the user requested so
    if args.plot:
        save_context_plots(ra, dec, targetname=targetname)


def K2findCampaigns_csv_main(args=None):
    """Exposes K2findCampaigns-csv to the command line."""
    parser = argparse.ArgumentParser(
                    description="Check which objects listed in a CSV table "
                                "are (or were) observable by NASA's K2 mission.")
    parser.add_argument('input_filename', nargs=1, type=str,
                        help="Path to a comma-separated table containing "
                             "columns 'ra,dec,kepmag' (decimal degrees) "
                             "or 'name'.")
    args = parser.parse_args(args)
    input_fn = args.input_filename[0]
    output_fn = input_fn + '-K2findCampaigns.csv'
    # First, try assuming the file has the classic "ra,dec,kepmag" format
    try:
        ra, dec, kepmag = parse_file(input_fn, exit_on_error=False)
        campaigns = np.array([findCampaigns(ra[idx], dec[idx])
                              for idx in range(len(ra))])
        output = np.array([ra, dec, kepmag, campaigns])
        print("Writing {0}.".format(output_fn))
        np.savetxt(output_fn, output.T, delimiter=', ',
                   fmt=['%10.10f', '%10.10f', '%10.2f', '%s'])
    # If this fails, assume the file has a single "name" column
    except ValueError:
        names = [name.strip() for name in open(input_fn, "r").readlines()
                 if len(name.strip()) > 0]
        print("Writing {0}.".format(output_fn))
        output = open(output_fn, "w")
        for target in names:
            try:
                campaigns, ra, dec = findCampaignsByName(target)
            except ValueError:
                campaigns = []
            output.write("{0}, {1}\n".format(target, campaigns))
            output.flush()
        output.close()



"""
The code will accept a file containing a minimum of two columns RA Dec.
where RA and Dec are in decimal degrees
"""
from __future__ import division, print_function
import sys

from . import logger

# Try importing numpy
try:
    import numpy as np
except ImportError:
    logger.error('You need numpy installed')
    sys.exit(1)

# Try importing matplotlib
try:
    import matplotlib.pyplot as pl
    got_mpl = True
    params = {
                'axes.linewidth': 1.5,
                'axes.labelsize': 24,
                'font.family': 'sans-serif',
                'font.size': 22,
                'legend.fontsize': 14,
                'xtick.labelsize': 16,
                'ytick.labelsize': 16,
                'text.usetex': False,
             }
    pl.rcParams.update(params)
except ImportError:
    logger.warning('You need matplotlib installed to get a plot')
    got_mpl = False

from . import fields
from . import projection as proj
from . import DEFAULT_PADDING


def angSepVincenty(ra1, dec1, ra2, dec2):
    """
    Vincenty formula for distances on a sphere
    """
    ra1_rad = np.radians(ra1)
    dec1_rad = np.radians(dec1)
    ra2_rad = np.radians(ra2)
    dec2_rad = np.radians(dec2)

    sin_dec1, cos_dec1 = np.sin(dec1_rad), np.cos(dec1_rad)
    sin_dec2, cos_dec2 = np.sin(dec2_rad), np.cos(dec2_rad)
    delta_ra = ra2_rad - ra1_rad
    cos_delta_ra, sin_delta_ra = np.cos(delta_ra), np.sin(delta_ra)

    diffpos = np.arctan2(np.sqrt((cos_dec2 * sin_delta_ra) ** 2 +
                         (cos_dec1 * sin_dec2 -
                          sin_dec1 * cos_dec2 * cos_delta_ra) ** 2),
                          sin_dec1 * sin_dec2 + cos_dec1 * cos_dec2 * cos_delta_ra)

    return np.degrees(diffpos)


def parse_file(infile, exit_on_error=True):
    """Parse a comma-separated file with columns "ra,dec,magnitude".
    """
    try:
        a, b, mag = np.atleast_2d(
                            np.genfromtxt(
                                        infile,
                                        usecols=[0, 1, 2],
                                        delimiter=','
                                        )
                    ).T
    except IOError as e:
        if exit_on_error:
            logger.error("There seems to be a problem with the input file, "
                         "the format should be: RA_degrees (J2000), Dec_degrees (J2000), "
                         "Magnitude. There should be no header, columns should be "
                         "separated by a comma")
            sys.exit(1)
        else:
            raise e
    return a, b, mag


def onSiliconCheck(ra_deg, dec_deg, FovObj, padding_pix=DEFAULT_PADDING):
    dist = angSepVincenty(FovObj.ra0_deg, FovObj.dec0_deg, ra_deg, dec_deg)
    if dist >= 90.:
        return False
    # padding_pix=3 means that objects less than 3 pixels off the edge of
    # a channel are counted inside, to account for inaccuracies in K2fov.
    return FovObj.isOnSilicon(ra_deg, dec_deg, padding_pix=padding_pix)


def nearSiliconCheck(ra_deg, dec_deg, FovObj, max_sep=8.2):
    dist = angSepVincenty(FovObj.ra0_deg, FovObj.dec0_deg, ra_deg, dec_deg)
    if dist <= max_sep:
        return True
    else:
        return False


def getRaDecRollFromFieldnum(fieldnum):
    """Returns ra, dec, and roll for a campaign.

    All values returned are in decimal degrees.
    """
    info = fields.getFieldInfo(fieldnum)
    return (info["ra"], info["dec"], info["roll"])


def K2onSilicon(infile, fieldnum, do_nearSiliconCheck=False):
    """Checks whether targets are on silicon during a given campaign.

    This function will write a csv table called targets_siliconFlag.csv,
    which details the silicon status for each target listed in `infile`
    (0 = not on silicon, 2 = on silion).

    Parameters
    ----------
    infile : str
        Path to a csv table with columns ra_deg,dec_deg,magnitude (no header).

    fieldnum : int
        K2 Campaign number.

    do_nearSiliconCheck : bool
        If `True`, targets near (but not on) silicon are flagged with a "1".
    """
    ra_sources_deg, dec_sources_deg, mag = parse_file(infile)
    n_sources = np.shape(ra_sources_deg)[0]
    if n_sources > 500:
        logger.warning("Warning: there are {0} sources in your target list, "
                       "this could take some time".format(n_sources))

    k = fields.getKeplerFov(fieldnum)
    raDec = k.getCoordsOfChannelCorners()

    onSilicon = list(
                    map(
                        onSiliconCheck,
                        ra_sources_deg,
                        dec_sources_deg,
                        np.repeat(k, len(ra_sources_deg))
                        )
                    )
    onSilicon = np.array(onSilicon, dtype=bool)

    if do_nearSiliconCheck:
        nearSilicon = list(
                        map(
                            nearSiliconCheck,
                            ra_sources_deg,
                            dec_sources_deg,
                            np.repeat(k, len(ra_sources_deg))
                            )
                        )
        nearSilicon = np.array(nearSilicon, dtype=bool)

    if got_mpl:
        almost_black = '#262626'
        light_grey = np.array([float(248)/float(255)]*3)
        ph = proj.PlateCaree()
        k.plotPointing(ph, showOuts=False)
        targets = ph.skyToPix(ra_sources_deg, dec_sources_deg)
        targets = np.array(targets)
        fig = pl.gcf()
        ax = fig.gca()
        ax = fig.add_subplot(111)
        ax.scatter(*targets, color='#fc8d62', s=7, label='not on silicon')
        ax.scatter(targets[0][onSilicon], targets[1][onSilicon],
                   color='#66c2a5', s=8, label='on silicon')
        ax.set_xlabel('R.A. [degrees]', fontsize=16)
        ax.set_ylabel('Declination [degrees]', fontsize=16)
        ax.invert_xaxis()
        ax.minorticks_on()
        legend = ax.legend(loc=0, frameon=True, scatterpoints=1)
        rect = legend.get_frame()
        rect.set_alpha(0.3)
        rect.set_facecolor(light_grey)
        rect.set_linewidth(0.0)
        texts = legend.texts
        for t in texts:
            t.set_color(almost_black)
        fig.savefig('targets_fov.png', dpi=300)
        pl.close('all')

    # prints zero if target is not on silicon
    siliconFlag = np.zeros_like(ra_sources_deg)

    # print a 1 if target is near but not on silicon
    if do_nearSiliconCheck:
        siliconFlag = np.where(nearSilicon, 1, siliconFlag)

    # prints a 2 if target is on silicon
    siliconFlag = np.where(onSilicon, 2, siliconFlag)

    outarr = np.array([ra_sources_deg, dec_sources_deg, mag, siliconFlag])
    np.savetxt('targets_siliconFlag.csv', outarr.T, delimiter=', ',
               fmt=['%10.10f', '%10.10f', '%10.2f', '%i'])

    if got_mpl:
        print('I made two files: targets_siliconFlag.csv and targets_fov.png')
    else:
        print('I made one file: targets_siliconFlag.csv')


def K2onSilicon_main(args=None):
    """Function called when `K2onSilicon` is executed on the command line."""
    import argparse
    parser = argparse.ArgumentParser(
        description="Run K2onSilicon to find which targets in a "
                    "list call on active silicon for a given K2 campaign.")
    parser.add_argument('csv_file', type=str,
                        help="Name of input csv file with targets, column are "
                             "Ra_degrees, Dec_degrees, Kepmag")
    parser.add_argument('campaign', type=int, help='K2 Campaign number')
    args = parser.parse_args(args)
    K2onSilicon(args.csv_file, args.campaign)


if __name__ == '__main__':
    K2onSilicon_main()



"""Functions that detail the K2C9 microlensing superstamp.
"""
from __future__ import print_function

import os
import json
import numpy as np

from . import PACKAGEDIR, logger, getKeplerFov

__all__ = ['inMicrolensRegion', 'pixelInMicrolensRegion']

# Load the JSON file that defines the C9 superstamp
SUPERSTAMP_FN = os.path.join(PACKAGEDIR, "data", "k2-c9-microlens-region.json")
SUPERSTAMP = json.load(open(SUPERSTAMP_FN))


def inMicrolensRegion_main(args=None):
    """Exposes K2visible to the command line."""
    import argparse
    parser = argparse.ArgumentParser(
                    description="Check if a celestial coordinate is "
                                "inside the K2C9 microlensing superstamp.")
    parser.add_argument('ra', nargs=1, type=float,
                        help="Right Ascension in decimal degrees (J2000).")
    parser.add_argument('dec', nargs=1, type=float,
                        help="Declination in decimal degrees (J2000).")
    args = parser.parse_args(args)
    if inMicrolensRegion(args.ra, args.dec):
        print("Yes! The coordinate is inside the K2C9 superstamp.")
    else:
        print("Sorry, the coordinate is NOT inside the K2C9 superstamp.")


def inMicrolensRegion(ra_deg, dec_deg):
    """Returns `True` if the given sky oordinate falls on the K2C9 superstamp.

    Parameters
    ----------
    ra_deg : float
        Right Ascension (J2000) in decimal degrees.

    dec_deg : float
        Declination (J2000) in decimal degrees.

    Returns
    -------
    onMicrolensRegion : bool
        `True` if the given coordinate is within the K2C9 microlens superstamp.
    """
    logger.warning("Warning: you are using a preliminary version "
                   "of the K2C9 superstamp!")
    fov = getKeplerFov(9)
    try:
        ch, col, row = fov.getChannelColRow(ra_deg, dec_deg,
                                            allowIllegalReturnValues=False)
        return pixelInMicrolensRegion(ch, col, row)
    except ValueError:
        return False


def pixelInMicrolensRegion(ch, col, row):
    """Returns `True` if the given pixel falls inside the K2C9 superstamp.

    The superstamp is used for microlensing experiment and is an almost
    contiguous area of 2.8e6 pixels.
    """
    try:
        vertices_col = SUPERSTAMP["channels"][str(int(ch))]["vertices_col"]
        vertices_row = SUPERSTAMP["channels"][str(int(ch))]["vertices_row"]
    except KeyError:  # Channel does not appear in file
        return False
    # The point is in one of 5 channels which constitute the superstamp
    # so check if it falls inside the polygon for this channel
    inside = isPointInsidePolygon(col, row, vertices_col, vertices_row)
    return inside


def isPointInsidePolygon(x, y, vertices_x, vertices_y):
    """Check if a given point is inside a polygon.

    Parameters vertices_x[] and vertices_y[] define the polygon.
    The number of array elements is equal to number of vertices of the polygon.
    This function works for convex and concave polygons.

    Parameters
    ----------
    vertices_x, vertices_y : lists or arrays of floats
        Vertices that define the polygon.

    x, y : float
        Coordinates of the point to check.

    Returns
    -------
    inside : bool
        `True` if the point is inside the polygon.
    """
    inside = False
    for i in range(len(vertices_x)):
        j = i - 1
        if ((vertices_x[i] > x) != (vertices_x[j] > x)):
            if (y < (x - vertices_x[i]) * (vertices_y[i] - vertices_y[j]) / (vertices_x[i] - vertices_x[j]) + vertices_y[i]):
                inside = not inside
    return inside


class C9FootprintPlot(object):
    """Create a plot showing the C9 footprint and superstamp.
    """
    def __init__(self, axes=None):
        if axes is None:
            import matplotlib.pyplot as pl
            params = {
                        'axes.linewidth': 1.,
                        'axes.labelsize': 20,
                        'font.family': 'sans-serif',
                        'font.size': 22,
                        'legend.fontsize': 14,
                        'xtick.labelsize': 16,
                        'ytick.labelsize': 16,
                        'text.usetex': False,
                     }
            pl.rcParams.update(params)
            self.fig = pl.figure(figsize=(8, 6))
            self.ax = self.fig.add_subplot(111)
        else:
            self.ax = axes

        self.ax.set_ylim([-30, -14])
        self.ax.set_xlim([280, 260])
        self.ax.set_xlabel("RA [deg]")
        self.ax.set_ylabel("Dec [deg]")

    def plot_outline(self):
        """Plots the coverage of both the channels and the C9 superstamp."""
        fov = getKeplerFov(9)
        # Plot the superstamp
        for ch in SUPERSTAMP["channels"]:
            v_col = SUPERSTAMP["channels"][ch]["vertices_col"]
            v_row = SUPERSTAMP["channels"][ch]["vertices_row"]
            radec = np.array([
                                fov.getRaDecForChannelColRow(int(ch),
                                                             v_col[idx],
                                                             v_row[idx])
                                for idx in range(len(v_col))
                              ])
            self.ax.fill(radec[:, 0], radec[:, 1],
                         lw=0, facecolor="#ff1111", zorder=100)

        # Plot all channel outlines
        corners = fov.getCoordsOfChannelCorners()
        for ch in np.arange(1, 85, dtype=int):
            if ch in fov.brokenChannels:
                continue  # certain channel are no longer used
            idx = np.where(corners[::, 2] == ch)
            mdl = int(corners[idx, 0][0][0])
            out = int(corners[idx, 1][0][0])
            ra = corners[idx, 3][0]
            dec = corners[idx, 4][0]
            self.ax.fill(np.concatenate((ra, ra[:1])), np.concatenate((dec, dec[:1])),
                         lw=0, facecolor="#bbbbbb", zorder=90)


def plot_c9(output_fn="c9.png"):
    p = C9FootprintPlot()
    p.plot_outline()
    p.fig.tight_layout()
    p.fig.savefig(output_fn)



#!/usr/bin/env python
import sys
import os

if "publish" in sys.argv[-1]:
    os.system("python setup.py sdist upload")
    sys.exit()

try:
    from setuptools import setup
    setup
except ImportError:
    from distutils.core import setup
    setup


# Command-line tools
entry_points = {'console_scripts': [
    'K2onSilicon = K2fov.K2onSilicon:K2onSilicon_main',
    'k2onsilicon = K2fov.K2onSilicon:K2onSilicon_main',
    'K2inMicrolensRegion = K2fov.c9:inMicrolensRegion_main',
    'k2inmicrolensregion = K2fov.c9:inMicrolensRegion_main',
    'K2findCampaigns = K2fov.K2findCampaigns:K2findCampaigns_main',
    'k2findcampaigns = K2fov.K2findCampaigns:K2findCampaigns_main',
    'K2findCampaigns-byname = K2fov.K2findCampaigns:K2findCampaigns_byname_main',
    'k2findcampaigns-byname = K2fov.K2findCampaigns:K2findCampaigns_byname_main',
    'K2findCampaigns-csv = K2fov.K2findCampaigns:K2findCampaigns_csv_main',
    'k2findcampaigns-csv = K2fov.K2findCampaigns:K2findCampaigns_csv_main'
]}

setup(name='K2fov',
      version='3.0.2',
      description='Find which targets are in the field of view of K2',
      author='Fergal Mullally, Tom Barclay, Geert Barentsen',
      author_email='tom@tombarclay.com',
      url='https://github.com/KeplerGO/K2fov',
      packages=['K2fov'],
      package_data={'K2fov': ['data/*.json']},
      install_requires=["numpy>=1.8"],
      entry_points=entry_points,
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering :: Astronomy",
          ],
      )
